# ===================================
# MIDI driving alto recorder.
# For now assume the hands begin in correct
# pose.  Also trying to set up a 'pose-puller',
# but seem to run into obstacles there......
# ===================================

# From numanoid: Wind in the trees, tracking cameara

from Tkinter import *
import poser
# ===================================

global NoteList
global Tempos
global InitFingers

# ===================================

scene = poser.Scene()
totalframes = scene.NumFrames()
# QPM = 120.0 # default Quarternotes Per Minute (not directly used)
TPQ = 0.0  # default Ticks Per Quarternote.
TPS = 0.0  # default Ticks Per Second.
MPQ = 0.0  # default MicrosecondsPerQuarternote.  Should get filled in by SET_TEMPO meta-event, not header.
FPS = float(scene.FramesPerSecond())

OurTrack=5 # Should make this a list-select

StartFrame = 1
EndFrame = totalframes

# Set up the figure for the Player, to be currently selected fig.
PlayerFig = scene.CurrentFigure()

# The recorder itself is selected by name.
RecActor = scene.Actor('Recorder')


# ===================================
# Lists and so on
channelVoiceMessages = {0x80:"NOTE_OFF",
                        0x90:"NOTE_ON",
                        0xA0:"POLYPHONIC_KEY_PRESSURE",
                        0xB0:"CONTROLLER_CHANGE",
                        0xC0:"PROGRAM_CHANGE",
                        0xD0:"CHANNEL_KEY_PRESSURE",
                        0xE0:"PITCH_BEND" }

channelModeMessages = {  0x78:"ALL_SOUND_OFF",
                         0x79:"RESET_ALL_CONTROLLERS",
                         0x7A:"LOCAL_CONTROL",
                         0x7B:"ALL_NOTES_OFF",
                         0x7C:"OMNI_MODE_OFF",
                         0x7D:"OMNI_MODE_ON",
                         0x7E:"MONO_MODE_ON",
                         0x7F:"POLY_MODE_ON", }

metaEvents =           {   0x00:"SEQUENCE_NUMBER",
                           0x01:"TEXT_EVENT",
                           0x02:"COPYRIGHT_NOTICE",
                           0x03:"SEQUENCE_TRACK_NAME",
                           0x04:"INSTRUMENT_NAME",
                           0x05:"LYRIC",
                           0x06:"MARKER",
                           0x07:"CUE_POINT",
                           0x20:"MIDI_CHANNEL_PREFIX",
                           0x21:"MIDI_PORT",
                           0x2F:"END_OF_TRACK",
                           0x51:"SET_TEMPO",
                           0x54:"SMTPE_OFFSET",
                           0x58:"TIME_SIGNATURE",
                           0x59:"KEY_SIGNATURE",
                           0x7F:"SEQUENCER_SPECIFIC_META_EVENT" }

# ==

Tempos={} # Key is time in internal MIDI units, val is the new MPQ
NoteList=[]
TrackNames={}  # So we can provide a listbox

# Holds the names and the initial position.  Start the dict with 0 in vals.
InitFingers= {'lThumb1':0.0,'lIndex1':0.0,'lMid1':0.0,'lRing1':0.0,'rIndex1':0.0,'rMid1':0.0,'rRing1':0.0,'rPinky1':0.0}


'''
 Try this way. Chart the pitches that have a fingering;
 the value for each key is a tuple in the fixed order
 Lthumb,Lindex,Lmid,Lring,Rindex,Rmid,Rring,Rpinky
 with 1 for finger down, 0 for finger up, .5 for halfway.
 The halfway applies only on the thumb (half-hold) and the last two, with double holes.

 Try this:  On each KeyDown, restore the initial pose, then immediately (in
 same frame) apply the up values.  Do nothing on KeyUP.

 Note Pitch Lthumb,    Lindex,Lmid,Lring,     Rindex,Rmid,Rring,Rpinky
  F4   65    1          1      1    1          1      1    1     1
  Fs4  66    1          1      1    1          1      1    1     0.5
  G4   67    1          1      1    1          1      1    1     0
  Gs4  68    1          1      1    1          1      1    0.5   0
  A4   69    1          1      1    1          1      1    0     0
  As4  70    1          1      1    1          1      0    1     1
  B4   71    1          1      1    1          0      1    1     0
  C5   72    1          1      1    1          0      0    0     0
  Cs5  73    1          1      1    0          1      1    0.5   0
  D5   74    1          1      1    0          0      0    0     0
  Ds5  75    1          1      0    1          1      0    0     0
  E5   76    1          1      0    0          0      0    0     0
  F5   77    1          0      0    0          0      0    0     0
  Fs5  78    0          1      1    0          0      0    0     0
  G5   79    0          0      1    0          0      0    0     0
  Gs5  80    0          0      1    1          1      1    1     0
  A5   81    0.5        1      1    1          1      1    0     0
  As5  82    0.5        1      1    1          1      0    1     0
  B5   83    0.5        1      1    1          0      1    0     0
  C6   84    0.5        1      1    1          0      1    0     0
  Cs6  85    0.5        1      1    0          1      0    0     0
  D6   86    0.5        1      1    0          0      0    0     0
  Ds6  87    0.5        1      1    0          0      1    1     0
  E6   88    0.5        1      1    0          1      1    0     0
'''
# ===================================
# Separate dicts for each of the moving fingers
# ===================================

Pitches = [65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88]

lThumbFingerings = { 65: 1.0, 66: 1.0, 67: 1.0, 68: 1.0, 69: 1.0,
                     70: 1.0, 71: 1.0, 72: 1.0, 73: 1.0, 74: 1.0,
                     75: 1.0, 76: 1.0, 77: 1.0, 78: 0.0, 79: 0.0,
                     80: 0.0, 81: 0.5, 82: 0.5, 83: 0.5, 84: 0.5,
                     85: 0.5, 86: 0.5, 87: 0.5, 88: 0.5 }


# ==

lIndexFingerings = { 65: 1.0, 66: 1.0, 67: 1.0, 68: 1.0, 69: 1.0,
                     70: 1.0, 71: 1.0, 72: 1.0, 73: 1.0, 74: 1.0,
                     75: 1.0, 76: 1.0, 77: 0.0, 78: 1.0, 79: 0.0,
                     80: 0.0, 81: 1.0, 82: 1.0, 83: 1.0, 84: 1.0,
                     85: 1.0, 86: 1.0, 87: 1.0, 88: 1.0 }

# ==

lMidFingerings  =  { 65: 1.0, 66: 1.0, 67: 1.0, 68: 1.0, 69: 1.0,
                     70: 1.0, 71: 1.0, 72: 1.0, 73: 1.0, 74: 1.0,
                     75: 0.0, 76: 0.0, 77: 0.0, 78: 1.0, 79: 1.0,
                     80: 1.0, 81: 1.0, 82: 1.0, 83: 1.0, 84: 1.0,
                     85: 1.0, 86: 1.0, 87: 1.0, 88: 1.0 }

# ==

lRingFingerings =  { 65: 1.0, 66: 1.0, 67: 1.0, 68: 1.0, 69: 1.0,
                     70: 1.0, 71: 1.0, 72: 1.0, 73: 0.0, 74: 0.0,
                     75: 1.0, 76: 0.0, 77: 0.0, 78: 0.0, 79: 0.0,
                     80: 1.0, 81: 1.0, 82: 1.0, 83: 1.0, 84: 1.0,
                     85: 0.0, 86: 0.0, 87: 0.0, 88: 0.0 }

# ==

rIndexFingerings = { 65: 1.0, 66: 1.0, 67: 1.0, 68: 1.0, 69: 1.0,
                     70: 1.0, 71: 0.0, 72: 0.0, 73: 1.0, 74: 0.0,
                     75: 1.0, 76: 0.0, 77: 0.0, 78: 0.0, 79: 0.0,
                     80: 1.0, 81: 1.0, 82: 1.0, 83: 0.0, 84: 0.0,
                     85: 1.0, 86: 0.0, 87: 0.0, 88: 1.0 }

# ==

rMidFingerings   = { 65: 1.0, 66: 1.0, 67: 1.0, 68: 1.0, 69: 1.0,
                     70: 0.0, 71: 1.0, 72: 0.0, 73: 1.0, 74: 0.0,
                     75: 0.0, 76: 0.0, 77: 0.0, 78: 0.0, 79: 0.0,
                     80: 1.0, 81: 1.0, 82: 0.0, 83: 1.0, 84: 1.0,
                     85: 0.0, 86: 0.0, 87: 1.0, 88: 1.0}

# ==

rRingFingerings  = { 65: 1.0, 66: 1.0, 67: 1.0, 68: 0.5, 69: 0.0,
                     70: 1.0, 71: 1.0, 72: 0.0, 73: 0.5, 74: 0.0,
                     75: 0.0, 76: 0.0, 77: 0.0, 78: 0.0, 79: 0.0,
                     80: 1.0, 81: 0.0, 82: 1.0, 83: 0.0, 84: 0.0,
                     85: 0.0, 86: 0.0, 87: 1.0, 88: 0.0 }

# ==

rPinkyFingerings = { 65: 1.0, 66: 0.5, 67: 0.0, 68: 0.0, 69: 0.0,
                     70: 1.0, 71: 0.0, 72: 0.0, 73: 0.0, 74: 0.0,
                     75: 0.0, 76: 0.0, 77: 0.0, 78: 0.0, 79: 0.0,
                     80: 0.0, 81: 0.0, 82: 0.0, 83: 0.0, 84: 0.0,
                     85: 0.0, 86: 0.0, 87: 0.0, 88: 0.0 }


# ===================================
# Next section is based closely on:
# midi.py -- MIDI classes and parser in Python
# Placed into the public domain in December 2001 by Will Ware
# ===================================

def getNumber(Str, length):
    # MIDI uses big-endian for everything
    sum = 0
    for i in range(length):
        sum = (sum << 8) + ord(Str[i])
    return sum, Str[length:]

def getVariableLengthNumber(Str):
    sum = 0
    i = 0
    while 1:
        x = ord(Str[i])
        i = i + 1
        sum = (sum << 7) + (x & 0x7F)
        if not (x & 0x80):
            return sum, Str[i:]

runningStatus = None

class MidiEvent:

    def __init__(self, track):
        self.track = track
        self.time = None
        self.channel = self.pitch = self.velocity = self.data = None

    def __cmp__(self, other):
        # assert self.time != None and other.time != None
        return cmp(self.time, other.time)

    def __repr__(self):
        r = ("<MidiEvent %s, t=%s, track=%s, channel=%s" %
             (self.type,
              repr(self.time),
              self.track.index,
              repr(self.channel)))
        for attrib in ["pitch", "data", "velocity"]:
            if getattr(self, attrib) != None:
                r = r + ", " + attrib + "=" + repr(getattr(self, attrib))
        return r + ">"

    def read(self, time, Str):
        global TPS,TPQ,MPQ
        global runningStatus
        self.time = time
        # do we need to use running status?
        if not (ord(Str[0]) & 0x80):
            Str = runningStatus + Str
        runningStatus = x = Str[0]
        x = ord(x)
        y = x & 0xF0
        z = ord(Str[1])

        if y in channelVoiceMessages.keys():
            self.channel = (x & 0x0F) + 1
            self.type = channelVoiceMessages[y]
            if (self.type == "PROGRAM_CHANGE" or
                self.type == "CHANNEL_KEY_PRESSURE"):
                self.data = z
                return Str[2:]
            else:
                self.pitch = z
                self.velocity = ord(Str[2])
                channel = self.track.channels[self.channel - 1]
                if (self.type == "NOTE_OFF" or
                    (self.velocity == 0 and self.type == "NOTE_ON")):
                    channel.noteOff(self.pitch, self.time)
                elif self.type == "NOTE_ON":
                    channel.noteOn(self.pitch, self.time, self.velocity)
                return Str[3:]

        elif y == 0xB0 and z in channelModeMessages.keys():
            self.channel = (x & 0x0F) + 1
            self.type = channelModeMessages[z]
            if self.type == "LOCAL_CONTROL":
                self.data = (ord(Str[2]) == 0x7F)
            elif self.type == "MONO_MODE_ON":
                self.data = ord(Str[2])
            return Str[3:]

        elif x == 0xF0 or x == 0xF7:
            self.type = {0xF0: "F0_SYSEX_EVENT",
                         0xF7: "F7_SYSEX_EVENT"}[x]
            length, Str = getVariableLengthNumber(str[1:])
            self.data = Str[:length]
            return Str[length:]

        elif x == 0xFF:
            if z not in metaEvents.keys():
               #print "Unknown meta event: FF %02X" % z
                sys.stdout.flush()
                raise "Unknown midi event type"
            self.type = metaEvents[z]
            length, Str = getVariableLengthNumber(Str[2:])
            self.data = Str[:length]
            if self.type == "SET_TEMPO":
              MPQ = float(ord(self.data[0]))*65536.0 + float(ord(self.data[1]))*256.0 + float(ord(self.data[2]))
              # DEGUG NEW Let's try making a separate array of tempos at time. See
              # how these can be incorporated in the real action.
              # Convert these to FramesPerTick here, so we can check FPT directly
              # during the play loop.
              if MPQ != 0.0:     # Tempo given
                 if TPQ != 0.0:
                    FPT =  FPS / (1000000.0 * TPQ / MPQ)
                 else:
                    FPT =  FPS / (1000000.0 * 125000.0 / MPQ)
              else:  # use SMPTE method
                 FPT = FPS / TPS
              Tempos[self.time]=FPT

            return Str[length:]

        raise "Unknown midi event type"

class MidiChannel:

    """A channel (together with a track) provides the continuity connecting
    a NOTE_ON event with its corresponding NOTE_OFF event. Together, those
    define the beginning and ending times for a Note."""

    def __init__(self, track, index):
        self.index = index
        self.track = track
        self.pitches = { }

    def __repr__(self):
        return "<MIDI channel %d>" % self.index

    def noteOn(self, pitch, time, velocity):
        self.pitches[pitch] = (time, velocity)

    def noteOff(self, pitch, time):
        global NoteList
        if pitch in self.pitches.keys():
            keyDownTime, velocity = self.pitches[pitch]
            # Here's where the note gets added to list
            NoteList.append( (self.track.index, self.index, pitch, velocity, keyDownTime, time) )
            del self.pitches[pitch]
        # The case where the pitch isn't in the dictionary is illegal,
        # I think, but we probably better just ignore it.


class DeltaTime(MidiEvent):

    type = "DeltaTime"

    def read(self, oldstr):
        self.time, newstr = getVariableLengthNumber(oldstr)
        return self.time, newstr


class MidiTrack:

    def __init__(self, index):
        self.index = index
        self.events = [ ]
        self.channels = [ ]
        self.length = 0
        for i in range(16):
            self.channels.append(MidiChannel(self, i+1))

    def read(self, Str):
        time = 0
        assert Str[:4] == "MTrk"
        length, Str = getNumber(Str[4:], 4)
        self.length = length
        mystr = Str[:length]
        remainder = Str[length:]
        while mystr:
            delta_t = DeltaTime(self)
            dt, mystr = delta_t.read(mystr)
            time = time + dt
            self.events.append(delta_t)
            e = MidiEvent(self)
            mystr = e.read(time, mystr)
            self.events.append(e)
        return remainder

    def __repr__(self):
        r = "<MidiTrack %d -- %d events\n" % (self.index, len(self.events))
        for e in self.events:
            r = r + "    " + `e` + "\n"
        return r + "  >"


class MidiFile:
    def __init__(self):
        self.file = None
        self.format = 1
        self.tracks = [ ]
        self.ticksPerQuarterNote = 125000.0  # default
        self.ticksPerSecond = 0.0

    def open(self, filename, attrib="rb"):
        if filename == None:
            if attrib in ["r", "rb"]:
                self.file = sys.stdin
            else:
                self.file = sys.stdout
        else:
            self.file = open(filename, attrib)

    def __repr__(self):
        r = "<MidiFile %d tracks\n" % len(self.tracks)
        for t in self.tracks:
            r = r + "  " + `t` + "\n"
        return r + ">"

    def close(self):
        self.file.close()

    def read(self):
        self.readstr(self.file.read())

    def readstr(self, Str):
        global TPS,TPQ,MPQ
        assert Str[:4] == "MThd"
        length, Str = getNumber(Str[4:], 4)
        assert length == 6
        format, Str = getNumber(Str, 2)
        self.format = format
        assert format == 0 or format == 1   # dunno how to handle 2
        numTracks, Str = getNumber(Str, 2)
        division, Str = getNumber(Str, 2)
        if division & 0x8000:  # 1 in hibit, use SMPTE system.
            framesPerSecond = -( (division >> 8) | (-128) )
            ticksPerFrame = division & 0xFF
            if ticksPerFrame == 29: ticksPerFrame = 30  # drop frame
            self.ticksPerSecond = ticksPerFrame * framesPerSecond
            # Fill in globals
            TPS = self.ticksPerSecond
            TPQ = 0.0
        else:  # 0 in hibit, use ticks per quarter.
            self.ticksPerQuarterNote = division & 0x7FFF
            # Fill in globals
            TPQ = self.ticksPerQuarterNote
            TPS = 0.0  # So we know to figure this way.
            # (We don't have Tempo (MPQ) right here, so we
            # can't complete the calculation.  The Playmidi
            # has to do that after the file is fully read,
            # so that MPQ is available from meta-event.)

        for i in range(numTracks):
            trk = MidiTrack(i)
            Str = trk.read(Str)
            self.tracks.append(trk)

# ===================================
# End stuff derived from Will Ware
# ===================================
# ===================================

def KeyDown(frame,pitch,tracknum):

    if (frame >= totalframes) or (pitch not in Pitches): return

    scene.SetFrame(frame)

    # Treat each finger separately.  A loop seems nicer, but this allows individual action.

    # The init position is always the base for each change.  Observing a professional
    # player on TV, it looks like all the movement is in joint 1 for all fingers.
    # Where the Fingerings dict has 1.0, leave init as is.
    # Where the Fingerings dict has 0.5, raise 6 degrees.
    # Where the Fingerings dict has 0.0, raise 12 degrees.
    FingerName = 'lThumb1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = lThumbFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'lIndex1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = lIndexFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'lMid1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = lMidFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'lRing1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = lRingFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'rIndex1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = rIndexFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'rMid1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = rMidFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'rRing1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = rRingFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))
    FingerName = 'rPinky1'
    Parm=PlayerFig.ActorByInternalName(FingerName).Parameter("Bend")
    Val = rPinkyFingerings[pitch]
    Parm.SetValue(InitFingers[FingerName]+(12.0 * (1.0 - Val)))

    # For Misaki, puff the cheeks in on KeyDown and out on KeyUp.
    Parm=PlayerFig.Actor("Head").Parameter("LCheekBulge")
    Parm.SetValue(0.2)
    Parm=PlayerFig.Actor("Head").Parameter("RCheekBulge")
    Parm.SetValue(0.2)

# ===================================
def KeyUp(frame,name):
    # For Misaki, puff the cheeks in on KeyDown and out on KeyUp.
    Parm=PlayerFig.Actor("Head").Parameter("LCheekBulge")
    Parm.SetValue(1.2)
    Parm=PlayerFig.Actor("Head").Parameter("RCheekBulge")
    Parm.SetValue(1.2)
    return

# ===================================
# ===================================
# PlayMain is the core of the animating action.
# ===================================

def PlayMain(MIDIfileArg, WAVEfileArg):
    global PlayerFigure, TPS,TPQ,MPQ,InitFingers
    m = MidiFile()
    try:
      m.open(MIDIfileArg)
    except:
      raise('MIDI File not found.')
    m.read()
    m.close()

    # Fetch the initial position.  Assume it has been set properly
    # before this.  (If I can make the 'puller' work, it will
    # go just before here; if not, just rely on PZ2 for pose.)
    for FingerName in InitFingers.keys():
        Finger = PlayerFig.ActorByInternalName(FingerName)
        Val=Finger.Parameter("Bend").Value()
        # Hold the initial setting of main (1) joint in dict by name
        InitFingers[FingerName]=Val
        # Set the range to linear interp.
        Finger.SetRangeLinear(StartFrame,EndFrame)

    # If a WAV file is given, insert it in the Poser scene.
    # If not, no error; just a soundless performance.
    if WAVEfileArg:
       try:
         scene.SetSound(WAVEfileArg)
       except:
         pass
       # If sound loaded OK:
       scene.SetSoundRange(0,totalframes)

    # Mathy part:
    # Use the frames / second of Poser scene, along with the
    # Ticks / Sec from the MIDI file, to get a constant for FramesPerTick
    # that can be multiplied by each tick to get a frame number.

    # MIDI files use two different systems.
    # If the file-reader has given us a non-zero MPQ (from tempo event), use it.
    # Otherwise, we get:
    # MPQ(Microseconds Per Quarter) and TPQ, which is Ticks per quarter.
    # So TPmicro = (T/Q * (Q/M) or (TPQ) / (MPQ).
    # TPS then becomes TPmicro * 1e6, or 1e6 * TPQ / MPQ.
    # Then frames / tick will be FPS / TPS,
    # which is FPS / (1E6 * TPQ / MPQ)


    scene.SetFrame(0)

    # DEGUG NEW Set the FPT initially to the 0 member of Tempos.
    # This should always be present, because there is always
    # a tempo event in the header part of the file.
    FPT = Tempos[0]
    NextTempoIndex=1

    # Now take the NoteList and apply it to our figure.
    # and apply it to our keys. This is looped on the NoteList,
    # not on the frames.
    f=0
    for NoteListItem in NoteList:
      # Each listitem contains:
      # self.track.index, self.index, pitch, velocity, keyDownTime, keyUptime
      # Pitch is the piano note.  Middle C is 60.  We use the conversion
      # factor FramesPerTick (FPT) to change MIDI time to our frames,
      # and we use the KeyList array to convert MIDI pitch to keynames.
      # The time given in NoteListItem[4] is key down instant;
      # the time given in NoteListItem[5] is key up instant.

      # DEGUG NEW  Check the keydown instant against the next Tempo value. If
      # there is no value, do nothing special; if the instant is >= the next
      # value, apply this new tempo change to FPT and increment the index for next.
      try:
          NewFPT=Tempos[NextTempoIndex]
      except:
          pass
      else:
          FPT=NewFPT
          NextTempoIndex=NextTempoIndex+1

      PitchIndex = NoteListItem[2]    # Pick up pitch
      TrackNum = NoteListItem[1]
      # Process the milliseconds into frame number, and move the keys.
      KeyDown((float(NoteListItem[4]) * FPT), PitchIndex, TrackNum)
      KeyUp  ((float(NoteListItem[5]) * FPT), TrackNum)

      # End main loop


# ===================================
# Next section is code for TKinter 'front'
# ===================================

class App:
    def __init__(self, master, textMessage):

        self.master = master
        Label(master, text="Name of MIDI file:" ).grid(row=0, sticky=W)
        Label(master, text="Name of WAV file:"  ).grid(row=1, sticky=W)
        self.TrackFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.TrackFrame.grid(row=2,column=0,sticky=N+S+E,pady=2)

        self.TrackLabel=Label(self.TrackFrame,text="Pick tracks to follow:").grid(row=0, column=1,sticky=N)
        self.TrackScroll=Scrollbar(self.TrackFrame, orient=VERTICAL)
        self.TrackScroll.grid(row=1, column=0,sticky=N+S+E)
        self.TrackListBox=Listbox(self.TrackFrame, height=15, selectmode=MULTIPLE,yscrollcommand=self.TrackScroll.set,exportselection=0)
        self.TrackListBox.grid( row=1, column=1)
        self.TrackScroll["command"]=self.TrackListBox.yview


        self.MIDInameEntry = Entry(master)
        self.WAVEnameEntry = Entry(master)
        self.MIDInameEntry.grid(row=0, column=2)
        self.WAVEnameEntry.grid(row=1, column=2)

        # buttons for Go, cancel.
        self.buttonGo = Button(master, text="Go", command=self.handleGo)
        self.buttonGo.grid(row=3, column=1)
        self.buttonCancel = Button(master, text="Cancel", command=self.quit)
        self.buttonCancel.grid(row=3, column=2)

    # - - - - - - - - - - - - - - - - - -

    def handleGo(self):
        # Fetch all the entry boxes:
        try:
          MIDIfileName = self.MIDInameEntry.get()
        except:
          raise "No MIDI given."
        try:
          WAVEfileName = self.WAVEnameEntry.get()
        except:
          pass  # No wave is not error; just soundless
        # The real action
        PlayMain(MIDIfileName,WAVEfileName)
        self.master.quit()

    # - - - - - - - - - - - - - - - - - -

    def quit(self):
        self.master.quit()

    # - - - - - - - - - - - - - - - - - -

# ===================================
# TK Definitions finished.  Activate the TK mainloop.
# ===================================

root = Tk()
app = App(root, "")
root.mainloop()

# ===================================

